from __future__ import annotations

from dataclasses import dataclass

from .contants import FanSpeed, FanSwing, OperatingMode, Power


@dataclass
class InteriorUnit:
    id: int
    name: str
    power: Power
    mode: OperatingMode
    requested_temperature: float
    humidity: int  # don't know the effect/working, is it current or requested, ?
    fan_speed: FanSpeed
    fan_swing: FanSwing
    room_temperature: float
    relative_temperature: float
    updated_at: int
    online: bool
    online_updated_at: int
    vendor: str
    model_id: str

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, InteriorUnit):
            return False

        return self.__dict__ == other.__dict__



from __future__ import annotations

import copy
import logging
import uuid
from dataclasses import dataclass
from types import TracebackType
from typing import Callable, Self

from . import api, notifications
from .api import CommandResponse
from .contants import (
    DEFAULT_REST_API_HOST,
    DEFAULT_STOMP_WEBSOCKET_HOST,
    DEFAULT_WAIT_DONE,
    FanSpeed,
    FanSwing,
    OperatingMode,
    Power,
    TemperatureUnit,
)
from .errors import CommandFailedException, IllegalStateException, InteriorUnitNotFoundException
from .interior_unit_models import InteriorUnit

logger = logging.getLogger(__name__)


@dataclass
class ConnectionInfo:
    auth_manager: api.AuthManager
    user_profile: api.UserProfile
    notification_socket: notifications.NotificationsWebsocket
    notification_subscription_id: uuid.UUID


class HitachiAirCloud:
    _email: str
    _password: str
    _default_wait_done: int
    _api_host: str
    _api_port: int

    _command_manager: api.CommandManager
    _connection_info: ConnectionInfo | None
    _interior_units: dict[int, InteriorUnit]

    on_change: Callable[[dict[int, tuple[InteriorUnit | None, InteriorUnit | None]]], None] | None

    def __init__(
        self,
        email: str,
        password: str,
        default_wait_done: int = DEFAULT_WAIT_DONE,
        api_host: str = DEFAULT_REST_API_HOST,
        api_port: int = 443,
        notification_host: str = DEFAULT_STOMP_WEBSOCKET_HOST,
    ) -> None:
        self._email = email
        self._password = password
        self._default_wait_done = default_wait_done
        self._api_host = api_host
        self._api_port = api_port
        self.notification_host = notification_host

        self._command_manager = api.CommandManager(self._get_auth_token_or_fail, host=api_host, port=api_port)
        self._connection_info = None
        self._interior_units = {}

        self.on_change = None

    async def __aenter__(self) -> Self:
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool:
        await self.close()
        return False

    @property
    def is_open(self) -> bool:
        return self._connection_info is not None

    @property
    def interior_units(self) -> list[InteriorUnit]:
        return [copy.deepcopy(iu) for iu in self._interior_units.values()]

    @property
    def temperature_unit(self) -> TemperatureUnit:
        if self._connection_info is None:
            raise Exception("AirCloud is not connected")

        if self._connection_info.user_profile.settings.temperatureUnit == "degC":
            return "CELSIUS"

        return "FAHRENHEIT"

    def find_interior_unit(self, rac_id: int) -> InteriorUnit | None:
        return self._interior_units.get(rac_id)

    def get_interior_unit(self, rac_id: int) -> InteriorUnit:
        iu = self.find_interior_unit(rac_id)
        if iu is None:
            raise InteriorUnitNotFoundException(f"Interior unit {rac_id} not found")
        return iu

    async def _get_auth_token_or_fail(self) -> str:
        if self._connection_info is None:
            raise Exception("AirCloud is not connected")
        return await self._connection_info.auth_manager.token()

    async def _wait_done(self, override_default_wait_done: int | None, command: CommandResponse) -> None:
        wait_done = self._default_wait_done if override_default_wait_done is None else override_default_wait_done
        if wait_done > 0:
            final_state = await self._command_manager.wait_ack(command, wait_done)
            if final_state != "DONE":
                raise CommandFailedException(
                    f"Command state is still not done (current: {final_state}) after delay {wait_done}"
                )

    async def connect(self) -> None:
        if self.is_open:
            raise IllegalStateException("AirCloud already connected")

        auth_manager = api.AuthManager(self._email, self._password, self._api_host, self._api_port)
        user_profile = await api.fetch_profile(auth_manager.token, self._api_host, self._api_port)
        self._interior_units = {
            iu.id: iu
            for iu in await api.get_interior_units(
                auth_manager.token, user_profile.familyId, self._api_host, self._api_port
            )
        }

        notification_socket = notifications.NotificationsWebsocket(
            self.notification_host,
            auth_manager.token,
            user_profile.id,
            user_profile.familyId,
            lambda ius: self._update_interior_units(ius),
        )
        await notification_socket.connect()
        notification_subscription_id = await notification_socket.subscribe()

        self._connection_info = ConnectionInfo(
            auth_manager,
            user_profile,
            notification_socket,
            notification_subscription_id,
        )

        logger.info("Connected")

    async def close(self) -> None:
        try:
            if self._connection_info is not None:
                await self._connection_info.notification_socket.close()
        finally:
            self._connection_info = None
            self._interior_units = {}
        logging.info("Closed")

    def _update_interior_units(self, interior_units: list[InteriorUnit]) -> None:
        logger.debug("Received interior units update: %s", interior_units)
        changes: dict[int, tuple[InteriorUnit | None, InteriorUnit | None]] = {}
        for iu in interior_units:
            old_iu = self._interior_units[iu.id]
            if old_iu != iu:
                changes[iu.id] = (old_iu, iu)
            self._interior_units[iu.id] = iu

        if self.on_change is not None:
            self.on_change(changes)

    async def update_all(self) -> None:
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")

        self._update_interior_units(
            await api.get_interior_units(
                self._connection_info.auth_manager.token,
                self._connection_info.user_profile.familyId,
                self._api_host,
                self._api_port,
            )
        )

    async def request_update_all(self) -> None:
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")

        await self._connection_info.notification_socket.refresh_all()

    async def request_update(self, rac_id: int) -> None:
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")

        await self._connection_info.notification_socket.refresh(rac_id)

    async def set(
        self,
        rac_id: int,
        power: Power | None = None,
        mode: OperatingMode | None = None,
        requested_temperature: float | None = None,
        humidity: int | None = None,
        fan_speed: FanSpeed | None = None,
        fan_swing: FanSwing | None = None,
        override_default_wait_done: int | None = None,
    ) -> None:
        """
        Send command to set interior_unit state

        :raises:
            IllegalStateException: If instance is not connected
            CommandFailedException: If state isn't done after wait delay
        """
        if self._connection_info is None:
            raise IllegalStateException("Connect must be called before calling this method")
        interior_unit = self.get_interior_unit(rac_id)

        command: CommandResponse = await api.configure_interior_unit(
            self._connection_info.auth_manager.token,
            self._connection_info.user_profile.familyId,
            interior_unit,
            power,
            mode,
            requested_temperature,
            humidity,
            fan_speed,
            fan_swing,
            host=self._api_host,
        )
        await self._wait_done(override_default_wait_done, command)
        await self.request_update(rac_id)



from __future__ import annotations

import logging

from aircloudy.contants import DEFAULT_REST_API_HOST, FanSpeed, FanSwing, OperatingMode, Power, TokenSupplier

from ..errors import TooManyRequestsException
from ..interior_unit_models import InteriorUnit
from .http_client import perform_request
from .rac_models import CommandResponse, CommandStatus, PowerAllResponse, _GeneralControlCommand, _InteriorUnitRest

logger = logging.getLogger(__name__)


async def get_interior_units(
    token_supplier: TokenSupplier, family_id: int, host: str = DEFAULT_REST_API_HOST, port: int = 443
) -> list[InteriorUnit]:
    logger.debug("Get interior units")
    response = await perform_request(
        "GET", f"/rac/ownership/groups/{family_id}/idu-list", token_supplier=token_supplier, host=host, port=port
    )

    if response.status != 200:
        raise Exception(f"Call failed (status={response.status} body={response.body}")

    return [_InteriorUnitRest(d).to_internal_representation() for d in response.body_as_json]


async def get_command_status(
    token_supplier: TokenSupplier, commands: list[CommandResponse], host: str = DEFAULT_REST_API_HOST, port: int = 443
) -> list[CommandStatus]:
    logger.debug("Get command status")

    response = await perform_request(
        "POST",
        "/rac/status/command",
        body=[c.__dict__ for c in commands],
        token_supplier=token_supplier,
        host=host,
        port=port,
    )

    return [CommandStatus(s) for s in response.body_as_json]


async def configure_interior_unit(
    token_supplier: TokenSupplier,
    family_id: int,
    interior_unit: InteriorUnit,
    power: Power | None = None,
    mode: OperatingMode | None = None,
    requested_temperature: float | None = None,
    humidity: int | None = None,
    fan_speed: FanSpeed | None = None,
    fan_swing: FanSwing | None = None,
    host: str = DEFAULT_REST_API_HOST,
    port: int = 443,
) -> CommandResponse:
    """Send command to change interior unit (like a remote control)

    :raises:
        TooManyRequestsException: If previous command is still in progress
    """
    command = _GeneralControlCommand(interior_unit, power, mode, requested_temperature, humidity, fan_speed, fan_swing)

    logger.debug("Configure interior unit familiy_id=%s : %s", family_id, command.__dict__)
    response = await perform_request(
        "PUT",
        f"/rac/basic-idu-control/general-control-command/{command.id}?familyId={family_id}",
        body=command.__dict__,
        do_not_raise_exception_on=(200, 429),
        token_supplier=token_supplier,
        host=host,
        port=port,
    )

    if response.status == 429:
        raise TooManyRequestsException(response.body)

    return CommandResponse(response.body_as_json)


async def request_refresh_interior_unit_state(
    token_supplier: TokenSupplier, rac_id: int, family_id: int, host: str = DEFAULT_REST_API_HOST, port: int = 443
) -> None:
    logger.debug("Request refresh interior unit state for rac id=%s, family_id=%s", rac_id, family_id)
    await perform_request(
        "PUT", f"/rac/status/{rac_id}?familyId={family_id}", token_supplier=token_supplier, host=host, port=port
    )


async def set_power(
    token_supplier: TokenSupplier, rac_id: str, power: Power, host: str = DEFAULT_REST_API_HOST, port: int = 443
) -> None:
    logger.debug("Set power rac_id=%s, power=%s", rac_id, power)
    await perform_request(
        "PUT",
        f"/rac/basic-idu-control/switch-on-off/{rac_id}",
        body={
            "power": power,
        },
        token_supplier=token_supplier,
        host=host,
        port=port,
    )


async def set_power_all(
    token_supplier: TokenSupplier,
    family_id: str,
    power: Power,
    interior_units: list[InteriorUnit],
    host: str = DEFAULT_REST_API_HOST,
    port: int = 443,
) -> PowerAllResponse:
    match power:
        case "ON":
            url = f"/rac/manage-idu/groups/{family_id}/idu/start"
        case "OFF":
            url = f"/rac/manage-idu/groups/{family_id}/idu/stop"
        case _:
            raise Exception(f"Unknown power value {power}")

    units = [_GeneralControlCommand(iu, power=power).__dict__ for iu in interior_units]

    logger.debug("Set power all power=%s for %s", power, units)
    response = await perform_request(
        "PUT",
        url,
        body=units,
        do_not_raise_exception_on=(200, 207),
        token_supplier=token_supplier,
        host=host,
        port=port,
    )

    return PowerAllResponse(response.body_as_json)
